// SPDX-License-Identifier: MIT

#ifndef RGBDS_LINK_ASSIGN_HPP
#define RGBDS_LINK_ASSIGN_HPP

// Assigns all sections a slice of the address space
void assign_AssignSections();

#endif // RGBDS_LINK_ASSIGN_HPP
